/*
 ************************************************************************
 *                                                                      *
 * udpUtilities -- utility routine for UDPs                             *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2011/2019  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include "OpenCSM.h"

#define STRLEN(A)   (int)strlen(A)


/*
 ************************************************************************
 *                                                                      *
 *   udpErrorStr - print error string                                   *
 *                                                                      *
 ************************************************************************
 */

/*@null@*/ char *
udpErrorStr(int stat)                   /* (in)  status number */
{
    char *string;                       /* (out) error message */

    string = (char *) EG_alloc(25*sizeof(char));
    if (string == NULL) {
        return string;
    }
    snprintf(string, 25, "EGADS status = %d", stat);

    return string;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpInitialize - initialize and get info about the list of arguments*
 *                                                                      *
 ************************************************************************
 */

int
udpInitialize(int    *nArgs,            /* (out) number of arguments */
              char   **namex[],         /* (out) array  of argument names */
              int    *typex[],          /* (out) array  of argument types */
              int    *idefault[],       /* (out) array  of argument integer defaults */
              double *ddefault[])       /* (out) array  of argument double  defaults */
{
    int    iarg, ival, *pint;
    double *p1dbl, *p2dbl;
    char   *pchar;

#ifdef DEBUG
    printf("udpInitialize()\n");
#endif

    /* call the printEgo now to avoid compiler messages that it is unused */
    printEgo(NULL);

    /* make the initial array that holds the udps */
    if (udps == NULL) {

        udps = (udp_T *) EG_alloc(sizeof(udp_T));
        if (udps == NULL) return EGADS_MALLOC;
        udps[0].ebody = NULL;

        for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
            udps[0].arg[iarg].size = 1;

            if        (argTypes[iarg] == +ATTRSTRING) {
                udps[0].arg[iarg].val = (char   *) EG_alloc(MAX_EXPR_LEN*sizeof(char  ));
                udps[0].arg[iarg].dot = NULL;
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
            } else if (argTypes[iarg] == +ATTRINT) {
                udps[0].arg[iarg].val = (int    *) EG_alloc(  sizeof(int   ));
                udps[0].arg[iarg].dot = NULL;
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
            } else if (argTypes[iarg] == -ATTRINT) {
                udps[0].arg[iarg].val = (int    *) EG_alloc(  sizeof(int   ));
                udps[0].arg[iarg].dot = NULL;
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
            } else if (argTypes[iarg] == +ATTRREAL) {
                udps[0].arg[iarg].val = (double *) EG_alloc(  sizeof(double));
                udps[0].arg[iarg].dot = NULL;
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
            } else if (argTypes[iarg] == -ATTRREAL) {
                udps[0].arg[iarg].val = (double *) EG_alloc(  sizeof(double));
                udps[0].arg[iarg].dot = NULL;
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
            } else if (argTypes[iarg] == +ATTRREALSEN) {
                udps[0].arg[iarg].val = (double *) EG_alloc(  sizeof(double));
                udps[0].arg[iarg].dot = (double *) EG_alloc(  sizeof(double));
                if (udps[0].arg[iarg].val == NULL || udps[0].arg[iarg].dot == NULL) return EGADS_MALLOC;
            } else if (argTypes[iarg] == 0) {
                udps[0].arg[iarg].val = (double *) EG_alloc(  sizeof(double));
                udps[0].arg[iarg].dot = NULL;
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
            } else {
                printf("bad argType[%d]=%d in udpInitialize\n", iarg, argTypes[iarg]);
                exit(0);
            }
        }
    }

    /* initialize the array elements that will hold the "current" settings */
    udps[0].ebody = NULL;
    udps[0].data  = NULL;

    for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
        if        (argTypes[iarg] == +ATTRSTRING) {
            pchar  = (char   *) (udps[0].arg[iarg].val);

            strcpy(pchar, "");
        } else if (argTypes[iarg] == +ATTRINT) {
            udps[0].arg[iarg].val = (int *) EG_reall(udps[0].arg[iarg].val, udps[0].arg[iarg].size*sizeof(int));
            if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;

            pint  = (int    *) (udps[0].arg[iarg].val);

            for (ival = 0; ival < udps[0].arg[iarg].size; ival++) {
                pint[ival] = argIdefs[iarg];
            }
        } else if (argTypes[iarg] == +ATTRREAL) {
            udps[0].arg[iarg].val = (double *) EG_reall(udps[0].arg[iarg].val, udps[0].arg[iarg].size*sizeof(double));
            if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;

            p1dbl  = (double *) (udps[0].arg[iarg].val);

            for (ival = 0; ival < udps[0].arg[iarg].size; ival++) {
                p1dbl[ival] = argDdefs[iarg];
            }
        } else if (argTypes[iarg] == +ATTRREALSEN) {
            udps[0].arg[iarg].val = (double *) EG_reall(udps[0].arg[iarg].val, udps[0].arg[iarg].size*sizeof(double));
            udps[0].arg[iarg].dot = (double *) EG_reall(udps[0].arg[iarg].dot, udps[0].arg[iarg].size*sizeof(double));
            if (udps[0].arg[iarg].val == NULL || udps[0].arg[iarg].dot == NULL) return EGADS_MALLOC;

            p1dbl = (double *) (udps[0].arg[iarg].val);
            p2dbl = (double *) (udps[0].arg[iarg].dot);

            for (ival = 0; ival < udps[0].arg[iarg].size; ival++) {
                p1dbl[ival] = argDdefs[iarg];
                p2dbl[ival] = 0;
            }
        }
    }

    /* set up returns that describe the UDP */
    *nArgs    = NUMUDPARGS;
    *namex    = (char**)argNames;
    *typex    = argTypes;
    *idefault = argIdefs;
    *ddefault = argDdefs;

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpNumBodys - number of Bodys expected in first arg to udpExecute  *
 *                                                                      *
 ************************************************************************
 */

int
udpNumBodys()
{

#ifdef DEBUG
    printf("udpNumBodys()\n");
#endif

#ifdef NUMUDPINPUTBODYS
    return NUMUDPINPUTBODYS;
#else
    return 0;
#endif
}



/*
 ************************************************************************
 *                                                                      *
 *   udpReset - reset the arguments to their defaults                   *
 *                                                                      *
 ************************************************************************
 */

int
udpReset(int flag)                      /* (in)  flag: 0=reset current, 1=close up */
{
    int  iarg, ival, iudp;

#ifdef DEBUG
    printf("udpReset(flag=%d)\n", flag);
#endif

    /* reset the "current" settings */
    if (flag == 0) {
        for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
            udps[0].arg[iarg].size = 1;

            if        (argTypes[iarg] == +ATTRSTRING) {
                strcpy((char*)(udps[0].arg[iarg].val), "");
            } else if (argTypes[iarg] == +ATTRINT) {
                int    *p  = (int    *) (udps[0].arg[iarg].val);

                for (ival = 0; ival < udps[0].arg[iarg].size; ival++) {
                    p[ival] = argIdefs[iarg];
                }
            } else if (argTypes[iarg] == +ATTRREAL) {
                double *p  = (double *) (udps[0].arg[iarg].val);

                for (ival = 0; ival < udps[0].arg[iarg].size; ival++) {
                    p[ival] = argDdefs[iarg];
                }
            } else if (argTypes[iarg] == +ATTRREALSEN) {
                double *p1 = (double *) (udps[0].arg[iarg].val);
                double *p2 = (double *) (udps[0].arg[iarg].dot);

                for (ival = 0; ival < udps[0].arg[iarg].size; ival++) {
                    p1[ival] = argDdefs[iarg];
                    p2[ival] = 0;
                }
            }
        }

        /* reset all the velocities */
        for (iudp = 0; iudp <= numUdp; iudp++) {
            for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
                if (argTypes[iarg] == +ATTRREALSEN) {
                    double *p2 = (double *) (udps[iudp].arg[iarg].dot);

                    for (ival = 0; ival < udps[iudp].arg[iarg].size; ival++) {
                        p2[ival] = 0;
                    }
                }
            }
        }

    /* called when closing up */
    } else if (udps != NULL) {
        for (iudp = 0; iudp <= numUdp; iudp++) {

            /* arguments */
            for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
                if (udps[iudp].arg[iarg].val != NULL) {
                    EG_free(udps[iudp].arg[iarg].val);
                }

                if (udps[iudp].arg[iarg].dot != NULL) {
                    EG_free(udps[iudp].arg[iarg].dot);
                }
            }

            /* private data */
            if (udps[iudp].data != NULL) {
#ifdef FREEUDPDATA
                FREEUDPDATA(udps[iudp].data);
#else
                EG_free(udps[iudp].data);
#endif
                udps[iudp].data = NULL;
            }
        }

        EG_free(udps);
        udps   = NULL;
        numUdp = 0;
    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSet - set an argument                                           *
 *                                                                      *
 ************************************************************************
 */

int
udpSet(char name[],                     /* (in)  argument name */
       void *value,                     /* (in)  pointer to values (can be char* or double*) */
       int  nvalue)                     /* (in)  number  of values */
{
    int  i, iarg, ivalue;
    char lowername[257];

#ifdef DEBUG
    printf("udpSet(name=%s, nvalue=%d)\n", name, nvalue);
#endif

    if        (name  == NULL) {
        return EGADS_NONAME;
    } else if (STRLEN(name) == 0 || STRLEN(name) > 255) {
        return EGADS_NONAME;
    } else if (value == NULL) {
        return EGADS_NULLOBJ;
    }

    for (i = 0; i <= STRLEN(name); i++) {
        lowername[i] = tolower(name[i]);
    }

    for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
        if (strcmp(lowername, argNames[iarg]) == 0) {
            udps[0].arg[iarg].size = nvalue;

            if        (argTypes[iarg] == +ATTRSTRING) {
                char   *p;
                char   *valueP = (char   *) value;

                #ifdef DEBUG
                    printf("   value=%s\n", valueP);
                #endif

                udps[0].arg[iarg].val = (char *) EG_reall(udps[0].arg[iarg].val, (nvalue+1)*sizeof(char  ));
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
                p = (char   *) (udps[0].arg[iarg].val);

                memcpy(p, valueP, (nvalue+1)*sizeof(char));
                return EGADS_SUCCESS;
            } else if (argTypes[iarg] == +ATTRINT) {
                int    *p;
                double *valueP = (double *) value;

                #ifdef DEBUG
                    printf("   value=%f", valueP[0]);
                    for (ivalue = 1; ivalue < nvalue; ivalue++) {
                        printf(" %f", valueP[ivalue]);
                    }
                    printf("\n");
                #endif

                udps[0].arg[iarg].val = (int   *) EG_reall(udps[0].arg[iarg].val, nvalue*sizeof(int   ));
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
                p = (int    *) (udps[0].arg[iarg].val);

                for (ivalue = 0; ivalue < nvalue; ivalue++) {
                    p[ivalue] = NINT(valueP[ivalue]);
                }
                return EGADS_SUCCESS;
            } else if (argTypes[iarg] == +ATTRREAL) {
                double *p;
                double *valueP = (double *) value;

                #ifdef DEBUG
                    printf("   value=%f", valueP[0]);
                    for (ivalue = 1; ivalue < nvalue; ivalue++) {
                        printf(" %f", valueP[ivalue]);
                    }
                    printf("\n");
                #endif

                udps[0].arg[iarg].val = (double *) EG_reall(udps[0].arg[iarg].val, nvalue*sizeof(double));
                if (udps[0].arg[iarg].val == NULL) return EGADS_MALLOC;
                p = (double *) (udps[0].arg[iarg].val);

                for (ivalue = 0; ivalue < nvalue; ivalue++) {
                    p[ivalue] = valueP[ivalue];
                }
                return EGADS_SUCCESS;
            } else if (argTypes[iarg] == +ATTRREALSEN) {
                double *p1, *p2;
                double *valueP = (double *) value;

                #ifdef DEBUG
                    printf("   value=%f", valueP[0]);
                    for (ivalue = 1; ivalue < nvalue; ivalue++) {
                        printf(" %f", valueP[ivalue]);
                    }
                    printf("\n");
                #endif

                udps[0].arg[iarg].val = (double *) EG_reall(udps[0].arg[iarg].val, nvalue*sizeof(double));
                udps[0].arg[iarg].dot = (double *) EG_reall(udps[0].arg[iarg].dot, nvalue*sizeof(double));
                if (udps[0].arg[iarg].val == NULL || udps[0].arg[iarg].dot == NULL) return EGADS_MALLOC;
                p1 = (double *) (udps[0].arg[iarg].val);
                p2 = (double *) (udps[0].arg[iarg].dot);

                for (ivalue = 0; ivalue < nvalue; ivalue++) {
                    p1[ivalue] = valueP[ivalue];
                    p2[ivalue] = 0;
                }
                return EGADS_SUCCESS;
            }
        }
    }

    printf(" udpSet: Parameter \"%s\" not known.  should be one of:", name);
    for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
        if (argTypes[iarg] == +ATTRSTRING ||
            argTypes[iarg] == +ATTRINT    ||
            argTypes[iarg] == +ATTRREAL   ||
            argTypes[iarg] == +ATTRREALSEN  ) {
            printf(" %s", argNames[iarg]);
        }
    }
    printf("\n");
    return EGADS_INDEXERR;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpGet - return an output parameter                                *
 *                                                                      *
 ************************************************************************
 */

int
udpGet(ego    ebody,                    /* (in)  Body pointer */
       char   name[],                   /* (in)  argument name */
       void   *value)                   /* (out) argument value (can be int* or double*) */
{
    int  i, iudp, judp, iarg;
    char lowername[257];

#ifdef DEBUG
    printf("udpGet(ebody=%llx, name=%s)\n", (long long)ebody, name);
#endif

    if (name == NULL) {
        return EGADS_NONAME;
    } else if (STRLEN(name) == 0 || STRLEN(name) > 255) {
        return EGADS_NONAME;
    }

    for (i = 0; i <= STRLEN(name); i++) {
        lowername[i] = tolower(name[i]);
    }

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
        if (strcmp(lowername, argNames[iarg]) == 0) {
            if        (argTypes[iarg] == -ATTRINT) {
                int    *p      = (int    *) (udps[0].arg[iarg].val);
                int    *valueP = (int    *) value;

                *valueP = p[0];
                return EGADS_SUCCESS;
            } else if (argTypes[iarg] == -ATTRREAL) {
                double *p      = (double *) (udps[0].arg[iarg].val);
                double *valueP = (double *) value;

                *valueP = p[0];
                return EGADS_SUCCESS;
            }
        }
    }

    printf(" udpGet: Parameter \"%s\" not known.  should be one of:", name);
    for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
        if (argTypes[iarg] == -ATTRINT ||
            argTypes[iarg] == -ATTRREAL  ) {
            printf(" %s", argNames[iarg]);
        }
    }
    printf("\n");
    return EGADS_INDEXERR;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpVel - set velocity of an argument                               *
 *                                                                      *
 ************************************************************************
 */

int
udpVel(ego    ebody,                    /* (in)  Body pointer */
       char   name[],                   /* (in)  argument name */
       double dot[],                    /* (in)  argument velocity(s) */
       int    ndot)                     /* (in)  number of velocities */
{
    int  i, iudp, judp, iarg, idot, hasdots;
    char lowername[257];

#ifdef DEBUG
    printf("udpVel(ebody=%llx, name=%s, dot=%f", (long long)ebody, name, dot[0]);
    for (i = 1; i < ndot; i++) {
        printf(" %f", dot[i]);
    }
    printf(")\n");
#endif

    if        (name  == NULL) {
        return EGADS_NONAME;
    } else if (STRLEN(name) == 0 || STRLEN(name) > 255) {
        return EGADS_NONAME;
    } else if (dot == NULL) {
        return EGADS_NULLOBJ;
    }

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    for (i = 0; i <= STRLEN(name); i++) {
        lowername[i] = tolower(name[i]);
    }

    for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
        if (strcmp(lowername, argNames[iarg]) == 0) {
            if (argTypes[iarg] == +ATTRREALSEN) {
                double *p = (double *) (udps[iudp].arg[iarg].dot);

                if (ndot != udps[iudp].arg[iarg].size) {
                    printf("size of value and dot do not agree\n");
                    return EGADS_RANGERR;
                }

                for (idot = 0; idot < ndot; idot++) {
                    p[idot] = dot[idot];
                }
                return EGADS_SUCCESS;
//$$$            } else if (argTypes[iarg] == +ATTRREAL) {
//$$$                double *p = (double *) (udps[iudp].arg[iarg].dot);
//$$$
//$$$                if (ndot != udps[iudp].arg[iarg].size) {
//$$$                    printf("size of value and dot do not agree\n");
//$$$                    return EGADS_RANGERR;
//$$$                }
//$$$
//$$$                for (idot = 0; idot < ndot; idot++) {
//$$$                    p[idot] = dot[idot];
//$$$                }
//$$$                return EGADS_NOTFOUND;
            } else {
                hasdots = 0;
                for (idot = 0; idot < ndot; idot++) {
                    if (fabs(dot[idot]) > 1.0e-6) {
                        hasdots++;
                    }
                }
                if (hasdots == 0) {
                    return EGADS_SUCCESS;
                }
            }
        }
    }

    return EGADS_INDEXERR;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpClean - clean the udp cache                                     *
 *                                                                      *
 ************************************************************************
 */

int
udpClean(ego ebody)                     /* (in)   Body pointer to clean from cache */
{
    int iudp, judp, iarg;

#ifdef DEBUG
    printf("udpClean(ebody=%llx)\n", (long long)ebody);
#endif

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;

            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    /* remove ebody from the udp's cache */
    udps[iudp].ebody = NULL;

    /* remove any private data from udp's cache */
    if (udps[iudp].data != NULL) {
#ifdef FREEUDPDATA
        FREEUDPDATA(udps[iudp].data);
#else
        EG_free(udps[iudp].data);
#endif
        udps[iudp].data = NULL;
    }

    /* decrement number of udps based upon NULL ebody entries */
    while (numUdp > 0) {
        if (udps[numUdp].ebody == NULL) {

            /* arguments */
            for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
                if (udps[numUdp].arg[iarg].val != NULL) {
                    EG_free(udps[numUdp].arg[iarg].val);
                }

                if (udps[numUdp].arg[iarg].dot != NULL) {
                    EG_free(udps[numUdp].arg[iarg].dot);
                }
            }

            /* private data */
            if (udps[numUdp].data != NULL) {
#ifdef FREEUDPDATA
                FREEUDPDATA(udps[numUdp].data);
#else
                EG_free(udps[numUdp].data);
#endif
                udps[numUdp].data = NULL;
            }

            numUdp--;
        } else {
            break;
        }
    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpMesh - return mesh associated with primitive                    *
 *                                                                      *
 ************************************************************************
 */

int
udpMesh(ego    ebody,                   /* (in)  Body pointer for mesh */
/*@unused@*/int    iMesh,               /* (in)  mesh index */
        int    *imax,                   /* (out) i-dimension of mesh */
        int    *jmax,                   /* (out) j-dimension of mesh */
        int    *kmax,                   /* (out) k-dimension of mesh */
        double *mesh[])                 /* (out) array of mesh points */
{
    int    iudp, judp;

#ifdef DEBUG
    printf("udpMesh(ebody=%llx)\n", (long long)ebody);
#endif

    *imax = 0;
    *jmax = 0;
    *kmax = 0;
    *mesh = NULL;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   cacheUdp - cache arguments of current instance                     *
 *                                                                      *
 ************************************************************************
 */

static int
cacheUdp()
{
    int    iarg, isize;

#ifdef DEBUG
    printf("cacheUdp()\n");
#endif

    /* increment number of UDPs in the cache */
    numUdp++;

    /* increase the arrays to make room for the new UDP */
    udps = (udp_T *) EG_reall(udps, (numUdp+1)*sizeof(udp_T));
    if (udps == NULL) return EGADS_MALLOC;
    udps[numUdp].ebody = NULL;

    /* copy info from udps[0] into udps[numUdp] */
    udps[numUdp].data = udps[0].data;

    for (iarg = 0; iarg < NUMUDPARGS; iarg++) {
        isize = udps[numUdp].arg[iarg].size = udps[0].arg[iarg].size;

        if (argTypes[iarg] == +ATTRSTRING) {
            udps[numUdp].arg[iarg].val = (char *) EG_alloc((isize+1)*sizeof(char  ));
            udps[numUdp].arg[iarg].dot = NULL;
            if (udps[numUdp].arg[iarg].val == NULL) return EGADS_MALLOC;

            memcpy(udps[numUdp].arg[iarg].val, udps[0].arg[iarg].val, (isize+1)*sizeof(char  ));
        } else if (argTypes[iarg] == +ATTRINT ||
                   argTypes[iarg] == -ATTRINT   ) {
            if (isize > 0) {
                udps[numUdp].arg[iarg].val = (int *) EG_alloc(isize*sizeof(int   ));
                udps[numUdp].arg[iarg].dot = NULL;
                if (udps[numUdp].arg[iarg].val == NULL) return EGADS_MALLOC;

                memcpy(udps[numUdp].arg[iarg].val, udps[0].arg[iarg].val, isize*sizeof(int   ));
            } else {
                udps[numUdp].arg[iarg].val = NULL;
                udps[numUdp].arg[iarg].dot = NULL;
            }
        } else if (argTypes[iarg] == +ATTRREAL ||
                   argTypes[iarg] == -ATTRREAL   ) {
            if (isize > 0) {
                udps[numUdp].arg[iarg].val = (double *) EG_alloc(isize*sizeof(double));
                udps[numUdp].arg[iarg].dot = NULL;
                if (udps[numUdp].arg[iarg].val == NULL) return EGADS_MALLOC;

                memcpy(udps[numUdp].arg[iarg].val, udps[0].arg[iarg].val, isize*sizeof(double));
            } else {
                udps[numUdp].arg[iarg].val = NULL;
                udps[numUdp].arg[iarg].dot = NULL;
            }
        } else if (argTypes[iarg] == +ATTRREALSEN) {
            if (isize > 0) {
                udps[numUdp].arg[iarg].val = (double *) EG_alloc(isize*sizeof(double));
                if (udps[numUdp].arg[iarg].val == NULL) return EGADS_MALLOC;

                memcpy(udps[numUdp].arg[iarg].val, udps[0].arg[iarg].val, isize*sizeof(double));

                udps[numUdp].arg[iarg].dot = (double *) EG_alloc(isize*sizeof(double));
                if (udps[numUdp].arg[iarg].dot == NULL) return EGADS_MALLOC;

                memcpy(udps[numUdp].arg[iarg].dot, udps[0].arg[iarg].dot, isize*sizeof(double));
            } else {
                udps[numUdp].arg[iarg].val = NULL;
                udps[numUdp].arg[iarg].dot = NULL;
            }
        } else if (argTypes[iarg] == 0) {
            if (isize > 0) {
                udps[numUdp].arg[iarg].val = (double *) EG_alloc(isize*sizeof(double));
                udps[numUdp].arg[iarg].dot = NULL;
                if (udps[numUdp].arg[iarg].val == NULL) return EGADS_MALLOC;

                memcpy(udps[numUdp].arg[iarg].val, udps[0].arg[iarg].val, isize*sizeof(double));
            } else {
                udps[numUdp].arg[iarg].val = NULL;
                udps[numUdp].arg[iarg].dot = NULL;
            }
        }
    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   printEgo - print EGADS topology for 5 levels                       *
 *                                                                      *
 ************************************************************************
 */

static void
printEgo(/*@null@*/ego    obj)                    /* (in)  ego to start */
{
    ego    eref0,    eref1,    eref2,    eref3,    eref4,   eref5;
    int    oclass0,  oclass1,  oclass2,  oclass3,  oclass4, oclass5;
    int    mtype0,   mtype1,   mtype2,   mtype3,   mtype4,  mtype5;
    int    nchild0,  nchild1,  nchild2,  nchild3,  nchild4, nchild5;
    int    ichild0,  ichild1,  ichild2,  ichild3,  ichild4, i;
    double data0[4], data1[4], data2[4], data3[4], data4[4], data5[4];
    ego    *ebodys0, *ebodys1, *ebodys2, *ebodys3, *ebodys4, *ebodys5;
    int    *senses0, *senses1, *senses2, *senses3, *senses4, *senses5;
    int    status, count=0;
    ego    topref, prev, next, refobj, context, prev1, next1;

    /* names take from egadsTypes.h */
    const char *classname[27] = {"contxt",       "transform",  "tessellation",  "nil",
                                 "empty",        "reference",  "ERROR 6",       "ERROR 7",
                                 "ERROR 8",      "ERROR 9",    "pcurve",        "curve",
                                 "surface",      "ERROR 13",   "ERROR 14",      "ERROR 15",
                                 "ERROR 16",     "ERROR 17",   "ERROR 18",      "ERROR 19",
                                 "node",         "edge",       "loop",          "face",
                                 "shell",        "body",       "model"};

    /* pcurves and curves */
    const char *mtypename1[10] = {"ERROR 0",     "line",       "circle",        "ellipse",
                                  "parabola",    "hyperbola",  "trimmed",       "bezier",
                                  "bspline",     "offset"};

    /* surfaces */
    const char *mtypename2[12] = {"ERROR 0",     "plane",      "spherical",     "cylindrical",
                                  "revolution",  "toroidal",   "trimmed",       "bezier",
                                  "bspline",     "ERROR 9",    "conical",       "extrusion"};

    /* faces */
    const char *mtypename3[ 3] = {"sreverse",    "nomtype",    "sforward"};

    /* other topology */
    const char *mtypename4[10] = {"nomtype",     "onenode",    "twonode",       "open",
                                  "closed",      "degenerate", "wirebody",      "facebody",
                                  "sheetbody",   "solidbody"};

    char mtypename[32];


    /* --------------------------------------------------------------- */

    if (obj == NULL) {
        return;
    }

    status = EG_getContext(obj, &context);
    if (status < 0) {
        printf("EG_getContext -> status=%d\n", status);
    }

    status = EG_getTopology(obj, &eref0, &oclass0, &mtype0,
                            data0, &nchild0, &ebodys0, &senses0);
    if (status != EGADS_SUCCESS) {
        EG_getInfo(obj, &oclass0, &mtype0, &topref, &prev, &next);
        nchild0 = -1;
        eref0   = topref;
    }

    if (oclass0 != NODE) {
        if (oclass0 == PCURVE || oclass0 == CURVE) {
            strcpy(mtypename, mtypename1[mtype0]);
        } else if (oclass0 == SURFACE) {
            strcpy(mtypename, mtypename2[mtype0]);
        } else if (oclass0 == FACE) {
            strcpy(mtypename, mtypename3[mtype0+1]);
        } else if (oclass0 == EDGE  || oclass0 == LOOP ||
                   oclass0 == SHELL || oclass0 == BODY ||
                   oclass0 == MODEL                      ) {
            strcpy(mtypename, mtypename4[mtype0]);
        } else {
            strcpy(mtypename, "");
        }

        printf("oclass0=%3d (%s)  mtype0=%3d (%s)  obj=%llx,  eref0=%llx\n",
               oclass0, classname[oclass0], mtype0, mtypename, (long long)obj, (long long)eref0);

        if (oclass0 == LOOP) {
            printf("< senses=");
            for (i = 0; i < nchild0; i++) {
                printf("%2d ", senses0[i]);
            }
            printf("\n");
        }
    } else {
        printf("oclass0=%3d (%s)  mtype0=%3d,  obj=%llx,  data0=%20.10e %20.10e %20.10e\n",
               oclass0, classname[oclass0], mtype0, (long long)obj, data0[0], data0[1], data0[2]);
    }

    next = obj->tref;
    prev = NULL;
    while (next != NULL) {
        refobj = (ego)(next->attrs);
        if (refobj != context) {
            count++;
            EG_getInfo(refobj, &oclass1, &mtype1, &topref, &prev1, &next1);
            printf("< refcount=%2d, refobj=%llx  (oclass=%2d (%s) mtype=%2d)\n",
                   count, (long long)refobj, oclass1, classname[oclass1], mtype1);
        }
        prev = next;
        next = (ego)(prev->blind);
    }

    for (ichild0 = 0; ichild0 < nchild0; ichild0++) {
        status = EG_getTopology(ebodys0[ichild0], &eref1, &oclass1, &mtype1,
                                data1, &nchild1, &ebodys1, &senses1);
        if (status != EGADS_SUCCESS) {
            EG_getInfo(obj, &oclass1, &mtype1, &topref, &prev, &next);
            nchild1 = -1;
            eref1   = topref;
        }

        if (oclass1 != NODE) {
            if (oclass1 == PCURVE || oclass1 == CURVE) {
                strcpy(mtypename, mtypename1[mtype1]);
            } else if (oclass1 == SURFACE) {
                strcpy(mtypename, mtypename2[mtype1]);
            } else if (oclass1 == FACE) {
                strcpy(mtypename, mtypename3[mtype1+1]);
            } else if (oclass1 == EDGE  || oclass1 == LOOP ||
                       oclass1 == SHELL || oclass1 == BODY ||
                       oclass1 == MODEL                      ) {
                strcpy(mtypename, mtypename4[mtype1]);
            } else {
                strcpy(mtypename, "");
            }

            printf(". oclass1=%3d (%s)  mtype1=%3d (%s)  obj=%llx,  eref1=%llx\n",
                   oclass1, classname[oclass1], mtype1, mtypename, (long long)(ebodys0[ichild0]), (long long)eref1);

            if (oclass1 == LOOP) {
                printf(". < senses=");
                for (i = 0; i < nchild1; i++) {
                    printf("%2d ", senses1[i]);
                }
                printf("\n");
            }
        } else {
            printf(". oclass1=%3d (%s)  mtype1=%3d,  obj=%llx,  data1=%20.10e %20.10e %20.10e\n",
                   oclass1, classname[oclass1], mtype1, (long long)(ebodys0[ichild0]), data1[0], data1[1], data1[2]);
        }

        for (ichild1 = 0; ichild1 < nchild1; ichild1++) {
            status = EG_getTopology(ebodys1[ichild1], &eref2, &oclass2, &mtype2,
                                    data2, &nchild2, &ebodys2, &senses2);
            if (status != EGADS_SUCCESS) {
                EG_getInfo(obj, &oclass2, &mtype2, &topref, &prev, &next);
                nchild2 = -1;
                eref2   = topref;
            }

            if (oclass2 != NODE) {
                if (oclass2 == PCURVE || oclass2 == CURVE) {
                    strcpy(mtypename, mtypename1[mtype2]);
                } else if (oclass2 == SURFACE) {
                    strcpy(mtypename, mtypename2[mtype2]);
                } else if (oclass2 == FACE) {
                    strcpy(mtypename, mtypename3[mtype2+1]);
                } else if (oclass2 == EDGE  || oclass2 == LOOP ||
                           oclass2 == SHELL || oclass2 == BODY ||
                           oclass2 == MODEL                      ) {
                    strcpy(mtypename, mtypename4[mtype2]);
                } else {
                    strcpy(mtypename, "");
                }

                printf(". . oclass2=%3d (%s)  mtype2=%3d (%s)  obj=%llx,  eref2=%llx\n",
                       oclass2, classname[oclass2], mtype2, mtypename, (long long)(ebodys1[ichild1]), (long long)eref2);

                if (oclass2 == LOOP) {
                    printf(". . < senses=");
                    for (i=0; i < nchild2; i++) {
                        printf("%2d ", senses2[i]);
                    }
                    printf("\n");
                }
            } else {
                printf(". . oclass2=%3d (%s)  mtype2=%3d,  obj=%llx,  data2=%20.10e %20.10e %20.10e\n",
                       oclass2, classname[oclass2], mtype2, (long long)(ebodys1[ichild1]), data2[0], data2[1], data2[2]);
            }

            for (ichild2 = 0; ichild2 < nchild2; ichild2++) {
                status = EG_getTopology(ebodys2[ichild2], &eref3, &oclass3, &mtype3,
                               data3, &nchild3, &ebodys3, &senses3);
                if (status != EGADS_SUCCESS) {
                    EG_getInfo(obj, &oclass3, &mtype3, &topref, &prev, &next);
                    nchild3 = -1;
                    eref3   = topref;
                }

                if (oclass3 != NODE) {
                    if (oclass3 == PCURVE || oclass3 == CURVE) {
                        strcpy(mtypename, mtypename1[mtype3]);
                    } else if (oclass3 == SURFACE) {
                        strcpy(mtypename, mtypename2[mtype3]);
                    } else if (oclass3 == FACE) {
                        strcpy(mtypename, mtypename3[mtype3+1]);
                    } else if (oclass3 == EDGE  || oclass3 == LOOP ||
                               oclass3 == SHELL || oclass3 == BODY ||
                               oclass3 == MODEL                      ) {
                        strcpy(mtypename, mtypename4[mtype3]);
                    } else {
                        strcpy(mtypename, "");
                    }

                    printf(". . . oclass3=%3d (%s)  mtype3=%3d (%s)  obj=%llx,  eref3=%llx\n",
                           oclass3, classname[oclass3], mtype3, mtypename, (long long)(ebodys2[ichild2]), (long long)eref3);

                    if (oclass3 == LOOP) {
                        printf(". . . < senses=");
                        for (i=0; i < nchild3; i++) {
                            printf("%2d ", senses3[i]);
                        }
                        printf("\n");
                    }
                } else {
                    printf(". . . oclass3=%3d (%s)  mtype3=%3d,  obj=%llx,  data3=%20.10e %20.10e %20.10e\n",
                           oclass3, classname[oclass3], mtype3, (long long)(ebodys2[ichild2]), data3[0], data3[1], data3[2]);
                }

                for (ichild3 = 0; ichild3 < nchild3; ichild3++) {
                    status = EG_getTopology(ebodys3[ichild3], &eref4, &oclass4, &mtype4,
                                   data4, &nchild4, &ebodys4, &senses4);
                    if (status != EGADS_SUCCESS) {
                        EG_getInfo(obj, &oclass4, &mtype4, &topref, &prev, &next);
                        nchild4 = -1;
                        eref4   = topref;
                    }

                    if (oclass4 != NODE) {
                        if (oclass4 == PCURVE || oclass4 == CURVE) {
                            strcpy(mtypename, mtypename1[mtype4]);
                        } else if (oclass4 == SURFACE) {
                            strcpy(mtypename, mtypename2[mtype4]);
                        } else if (oclass4 == FACE) {
                            strcpy(mtypename, mtypename3[mtype4+1]);
                        } else if (oclass4 == EDGE  || oclass4 == LOOP ||
                                   oclass4 == SHELL || oclass4 == BODY ||
                                   oclass4 == MODEL                      ) {
                            strcpy(mtypename, mtypename4[mtype4]);
                        } else {
                            strcpy(mtypename, "");
                        }

                        printf(". . . . oclass4=%3d (%s)  mtype4=%3d (%s)  obj=%llx,  eref4=%llx\n",
                               oclass4, classname[oclass4], mtype4, mtypename, (long long)(ebodys3[ichild3]), (long long)eref4);

                        if (oclass4 == LOOP) {
                            printf(". . . . < senses=");
                            for (i=0; i < nchild4; i++) {
                                printf("%2d ", senses4[i]);
                            }
                            printf("\n");
                        }
                    } else {
                        printf(". . . . oclass4=%3d (%s)  mtype4=%3d,  obj=%llx,  data4=%20.10e %20.10e %20.10e\n",
                               oclass4, classname[oclass4], mtype4, (long long)(ebodys3[ichild3]), data4[0], data4[1], data4[2]);
                    }

                    for (ichild4 = 0; ichild4 < nchild4; ichild4++) {
                        status = EG_getTopology(ebodys4[ichild4], &eref5, &oclass5, &mtype5,
                                                data5, &nchild5, &ebodys5, &senses5);
                        if (status != EGADS_SUCCESS) {
                            EG_getInfo(obj, &oclass5, &mtype5, &topref, &prev, &next);
                            nchild5 = -1;
                            eref5   = topref;
                        }

                        if (oclass5 != NODE) {
                            if (oclass5 == PCURVE || oclass5 == CURVE) {
                                strcpy(mtypename, mtypename1[mtype5]);
                            } else if (oclass5 == SURFACE) {
                                strcpy(mtypename, mtypename2[mtype5]);
                            } else if (oclass5 == FACE) {
                                strcpy(mtypename, mtypename3[mtype5+1]);
                            } else if (oclass5 == EDGE  || oclass5 == LOOP ||
                                       oclass5 == SHELL || oclass5 == BODY ||
                                       oclass5 == MODEL                      ) {
                                strcpy(mtypename, mtypename4[mtype5]);
                            } else {
                                strcpy(mtypename, "");
                            }

                            printf(". . . . . oclass5=%3d (%s)  mtype5=%3d (%s)  obj=%llx,  eref5=%llx\n",
                                   oclass5, classname[oclass5], mtype5, mtypename, (long long)(ebodys4[ichild4]), (long long)eref5);

                            if (oclass5 == LOOP) {
                                printf(". . . . . < senses=");
                                for (i=0; i < nchild5; i++) {
                                    printf("%2d ", senses5[i]);
                                }
                                printf("\n");
                            }
                        } else {
                            printf(". . . . . oclass5=%3d (%s)  mtype5=%3d,  obj=%llx,  data5=%20.10e %20.10e %20.10e\n",
                                   oclass5, classname[oclass5], mtype5, (long long)(ebodys4[ichild4]), data5[0], data5[1], data5[2]);
                        }
                    }
                }
            }
        }
    }
}
